
//		Palette.cpp - display and set the console palette from the command line


#define UNICODE							1
#define _UNICODE						1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>
#include <conio.h>

#include "PlugIn.h"
#include "TakeCmd.h"

#define MACRO_EXP( arg )				#arg
#define MACRO_STR( arg )				MACRO_EXP( arg )


#define PLUGIN_NAME						L"Palette"

#define VER_MAJOR						0
#define VER_MINOR						87
#define VER_BUILD						4
#define VER_PATCH						1

#ifdef _WIN64
#define PLATFORM						L"x64"
#else
#define PLATFORM						L"x86"
#endif

#define MAX_ARG_LEN						8192
#define SHORT_BUF_LEN					1024
#define MAX_FILENAME_LEN				32768


#define PALETTE_FLAG_NO_ERRORS			0x0001
#define PALETTE_FLAG_ALLOW_INCOMPLETE	0x0002
#define PALETTE_FLAG_QUIET				0x0004
#define PALETTE_FLAG_ALLOW_OVERWRITE	0x0008

#define PALETTE_FILE_TYPE_UNKNOWN		0
#define PALETTE_FILE_TYPE_NATIVE		1
#define PALETTE_FILE_TYPE_INI			2
#define PALETTE_FILE_TYPE_XML			3

#define SPACE							0x20
#define ESC								0x1b


#define ERRHANDLE						GetStdHandle( STD_ERROR_HANDLE )


HINSTANCE MyDLLHandle					= NULL;

HWND ConsoleWindow						= NULL;
HANDLE ConsoleHandle					= NULL;

unsigned int TccVerMajor				= 0;
unsigned int TccVerMinor				= 0;



const wchar_t PluginFeatures[] = L"palette,paledit,@palette,@pickcolor";



const wchar_t PaletteHelpText[]	= L"Display or change the console palette.\n\
\n\
PALETTE /Cn:color[,color...] /D /F:n /I /Q /R:file /S:file /T /V /W \r\n\
\n\
\t/Cn:\tredefine a console color; n = 0 to 15\n\
\t/D\tload a built-in set of default colors\n\
\t/F:\toutput format\n\
\t/I\tallow incomplete palette files\n\
\t/Q\tquietly\n\
\t/R:\tread palette from a file\n\
\t/S:\tsave palette to a file\n\
\t/T\tdisplay a test pattern\n\
\t/V\tverbose\n\
\t/W\tno warning if console window is not visible\n\
\n\
Color values may be given as W3C color names, or three- or six-digit \n\
RGB values with a leading #.  If n is less than 15, you can specify\n\
multiple values separated by commas.\n";


const wchar_t F_PaletteHelpText[]	= L"@PALETTE[n,fmt] \n\
\n\
Returns the value of console color #n.\n\
\n\
fmt is the format to return:\n\
\n\
0\tsix-digit hex RGB  (default)\n\
1\tthree- or six-digit hex RGB\n\
2\tW3C color name, or six-digit hex RGB\n\
3\tW3C color name, three- or six-digit RGB\n\
4\tdecimal red,green,blue\n\
5\tCOLORREF\n";


const wchar_t F_PickColorHelpText[]		= L"@PICKCOLOR[color,fmt,title] \n\
\n\
Opens the color picker dialog and returns the \n\
user\u2019s chosen color, or 0 if the user cancels the \n\
dialog.  The initial color may be  a W3C color\n\
name, a three- or six-digit hex RGB with a\n\
leading #, or $n where n is a console color\n\
index 0-15.\n\
\n\
fmt is the format to return:\n\
\n\
0\tsix-digit hex RGB  (default)\n\
1\tthree- or six-digit hex RGB\n\
2\tW3C color name, or six-digit hex RGB\n\
3\tW3C color name, three- or six-digit RGB\n\
4\tdecimal red,green,blue\n\
5\tCOLORREF\n";


const wchar_t IniSection[]		= L"table";


const wchar_t IniValueNames[16][15]		= { L"dark_black", L"dark_blue", L"dark_green", L"dark_cyan",
	L"dark_red", L"dark_magenta", L"dark_yellow", L"dark_white",
	L"bright_black", L"bright_blue", L"bright_green", L"bright_cyan",
	L"bright_red", L"bright_magenta", L"bright_yellow", L"bright_white" };


const wchar_t XmlColorNames[16][15]		= { L"Ansi 0 Color", L"Ansi 4 Color", L"Ansi 2 Color", L"Ansi 6 Color",
	L"Ansi 1 Color", L"Ansi 5 Color", L"Ansi 3 Color", L"Ansi 7 Color",
	L"Ansi 8 Color", L"Ansi 12 Color", L"Ansi 10 Color", L"Ansi 14 Color",
	L"Ansi 9 Color", L"Ansi 13 Color", L"Ansi 11 Color", L"Ansi 15 Color" };


const wchar_t TccColorNames[16][15]		= { L"Black", L"Blue", L"Green", L"Cyan", L"Red", L"Magenta", L"Yellow", L"White",
											L"Bright Black", L"Bright Blue", L"Bright Green", L"Bright Cyan", L"Bright Red",
											L"Bright Magenta", L"Bright Yellow", L"Bright White" };


const wchar_t DefaultsFilename[]	= L"<defaults>";


const COLORREF DefaultColors[16] = { 0x000000, 0x990000, 0x009900, 0x999900, 0x000099, 0x990099, 0x006699, 0xcccccc,
	0x666666, 0xff3333, 0x00ff00, 0xffff00, 0x0000ff, 0xff00ff, 0x00ffff, 0xffffff };


const wchar_t FormatName[4][5]	= { L"", L"text", L"INI", L"XML" };



struct _SaveInfo {
	LPTSTR Name;
	LPTSTR Author;
	LPTSTR Url;
	LPTSTR License;
	WORD Attribs;
	unsigned int DlgFlags;
};



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



int GetErrorText( int err, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( !FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) )
		wsprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



int GetLastErrorText( LPTSTR outString )
{
	return GetErrorText( GetLastError(), outString );
}



int GetTccVersion()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"%_4ver";

	unsigned int major				= 0;
	unsigned int minor				= 0;
	wchar_t *sep1					= NULL;


	ExpandVariables( temp, 1 );

	major = wcstoul( temp, &sep1, 10 );

	if ( sep1 ) {
		if ( ( *sep1 == '.' ) || ( *sep1 == ',' ) ) {
			minor = wcstoul( sep1 + 1, NULL, 10 );
		}
	}

	if ( ( major < 10 ) || ( major > 99 ) )
		major = minor = 0;

	TccVerMajor	= major;
	TccVerMinor	= minor;
	return 0;
}



BOOL PipeShell()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"%[_transient]%[_pipe]";


	ExpandVariables( temp, 1 );

	if ( wcscmp( temp, L"00" ) )
		return TRUE;
	else
		return FALSE;
}



BOOL ModernTCmdTab()
{
	if ( TccVerMajor < 28 )
		return FALSE;


	wchar_t temp[SHORT_BUF_LEN];


	TakeCommandIPC( L"HVIEW", temp );

	if ( !iswdigit( temp[0] ) )
		return FALSE;

	return TRUE;
}



LPTSTR _FilenamePart( LPTSTR Filename, LPTSTR outFilename )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	BOOL FoundSlash			= FALSE;
	size_t EndOfPathPart	= -1;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) ) {
			EndOfPathPart = i;
			FoundSlash = TRUE;
		}
		else if ( ( Filename[i] == ':' ) && !FoundSlash )
			EndOfPathPart = i;
	}

	if ( EndOfPathPart == -1 ) {
		if ( outFilename ) {
			wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename );
			return outFilename;
		}
		else
			return Filename;
	}

	if ( outFilename ) {
		wcscpy_s( outFilename, MAX_FILENAME_LEN, Filename + EndOfPathPart + 1 );
		return outFilename;
	}
	else
		return Filename + EndOfPathPart + 1;
}



LPTSTR _ExtensionPart( LPTSTR Filename, LPTSTR outExt )
{
	if ( !Filename )
		return NULL;
	else if ( Filename[0] == '\0' )
		return NULL;


	LPTSTR Ext			= NULL;


	for ( size_t i = 0; Filename[i]; i++ ) {
		if ( ( Filename[i] == '/' ) || ( Filename[i] == '\\' ) || ( Filename[i] == ':' ) )
			Ext = NULL;
		else if ( Filename[i] == '.' )
			Ext = Filename + i;
	}

	if ( !Ext )
		return NULL;

	if ( outExt ) {
		wcscpy_s( outExt, MAX_FILENAME_LEN, Ext );
		return outExt;
	}

	return Ext;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion, LPCSTR SourceFile, LPCSTR SourceFileVer )
{
	wchar_t File[MAX_FILENAME_LEN]		= L"";
	wchar_t Ver[SHORT_BUF_LEN]			= L"";


	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion ) {
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

		if ( SourceFile && SourceFileVer ) {
			MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, SourceFile, -1, File, MAX_FILENAME_LEN );
			MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, SourceFileVer, -1, Ver, SHORT_BUF_LEN );
			Printf( L"\t\t\t\t%s  %s\r\n", _FilenamePart( File, NULL), Ver );
		}
	}

	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	if ( ( ch == ':' ) || ( ch == '=' ) )
		return TRUE;
	else
		return FALSE;
}



BOOL IsListSep( wchar_t ch )
{
	if ( ( ch == ',' ) || ( ch == ';' ) )
		return TRUE;
	else
		return FALSE;
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value	= 0;
	wchar_t *sep1		= NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( !EndChar && ( *sep1 != '\0' ) )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( !EndChar && ( *sep1 != '\0' ) )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



//	Strips off any leading or trailing spaces from a string.

int TrimLine( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	int FirstNonSpace = -1, LastNonSpace = -1;
	wchar_t ch = '\0';

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( !iswspace( ch ) ) {
			LastNonSpace = i;

			if ( FirstNonSpace == -1 )
				FirstNonSpace = i;
		}
	}

	if ( FirstNonSpace == -1 ) {
		String[0] = '\0';
		return 0;
	}

	String[LastNonSpace + 1] = '\0';

	if ( FirstNonSpace ) {
		int i = 0;

		do {
			ch = String[FirstNonSpace + i];
			String[i++] = ch;

		} while ( ch );
	}

	return 0;
}



HWND MyWindow()
{
	wchar_t temp[SHORT_BUF_LEN]		= L"";

	HWND hWnd			= NULL;


	TakeCommandIPC( L"HWND", temp );

	if ( iswdigit( temp[0] ) ) {
		hWnd = (HWND) wcstoul( temp, NULL, 10 );

		if ( IsWindow( hWnd ) )
			return hWnd;
	}

	return GetConsoleWindow();
}



#include "ParseArgs.cpp"
#include "w3ccolors.cpp"
#define MM_INTERNET_FEATURES
#include "mmfiles.cpp"
#include "BufferedOutput.cpp"
#include "ForceFGWin.cpp"



struct _FTW_Data {
	LPTSTR Title;
	unsigned int Flags;
	HWND FoundWindow;
};



int BruteForceRefresh()
{
	CONSOLE_SCREEN_BUFFER_INFO BufferInfo;

	unsigned int BufferWidth		= 0;
	unsigned int BufferHeight		= 0;
	unsigned int BufferSize			= 0;

	wchar_t *SaveChars				= NULL;
	WORD *SaveAttr					= NULL;
	wchar_t *FlashChars				= NULL;
	WORD *FlashAttr					= NULL;

	WORD Attr						= 0;
	COORD Coord						= { 0, 0 };
	DWORD Xferred					= 0;
	int rv							= 0;


	memset( &BufferInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFO ) );
	if ( !GetConsoleScreenBufferInfo( ConsoleHandle, &BufferInfo ) )
		return 0;

	BufferWidth		= BufferInfo.dwSize.X;
	BufferHeight	= BufferInfo.srWindow.Bottom - BufferInfo.srWindow.Top + 1;
	BufferSize		= BufferWidth * BufferHeight;
	Attr			= ( BufferInfo.wAttributes & 0x00f0 ) | ( ( BufferInfo.wAttributes & 0x00f0 ) >> 4 );
	Coord.X			= 0;
	Coord.Y			= BufferInfo.srWindow.Top;

	if ( !Attr )
		Attr = 0x01;


	SaveChars	= (wchar_t*) malloc( BufferSize * sizeof( wchar_t ) );
	SaveAttr	= (WORD*) malloc( BufferSize * sizeof( WORD ) );
	FlashChars	= (wchar_t*) malloc( BufferSize * sizeof( wchar_t ) );
	FlashAttr	= (WORD*) malloc( BufferSize * sizeof( WORD ) );

	if ( SaveChars && SaveAttr && FlashChars && FlashAttr ) {
		if ( ReadConsoleOutputCharacter( ConsoleHandle, SaveChars, BufferSize, Coord, &Xferred ) && ReadConsoleOutputAttribute( ConsoleHandle, SaveAttr, BufferSize, Coord, &Xferred ) ) {
			wmemset( FlashChars, 0xb7, BufferSize );
			wmemset( (wchar_t*) FlashAttr, Attr, BufferSize );
			WriteConsoleOutputAttribute( ConsoleHandle, FlashAttr, BufferSize, Coord, &Xferred );
			WriteConsoleOutputCharacter( ConsoleHandle, FlashChars, BufferSize, Coord, &Xferred );
			Sleep( 75 );

			WriteConsoleOutputCharacter( ConsoleHandle, SaveChars, BufferSize, Coord, &Xferred );
			WriteConsoleOutputAttribute( ConsoleHandle, SaveAttr, BufferSize, Coord, &Xferred );
			rv = 1;
		}
	}

	free( SaveChars );
	free( SaveAttr );
	free( FlashChars );
	free( FlashAttr );
	return rv;
}



BOOL CALLBACK _FindTabWindowProc( HWND hWnd, LPARAM lParam )
{
	_FTW_Data *Data					= (_FTW_Data*) lParam;
	wchar_t Title[SHORT_BUF_LEN]	= L"";


	GetWindowText( hWnd, Title, SHORT_BUF_LEN );

	if ( wcscmp( Title, Data->Title ) )
		return TRUE;

	Data->FoundWindow = hWnd;
	return FALSE;
}



HWND FindTabWindow( HWND TCmdWindow )
{
	if ( !TCmdWindow )
		return NULL;


	HWND MyConsole						= GetConsoleWindow();
	wchar_t OldTitle[SHORT_BUF_LEN]		= L"";
	wchar_t NewTitle[SHORT_BUF_LEN]		= L"";
	_FTW_Data Data;


	GetWindowText( MyConsole, OldTitle, SHORT_BUF_LEN );

	wsprintf( NewTitle, L"\u2039%08X-%04X\u203a", GetCurrentProcessId(), rand() );
	Sleep( 30 );
	SetWindowText( MyConsole, NewTitle );

	Data.Title			= NewTitle;
	Data.Flags			= 0;
	Data.FoundWindow	= NULL;


	for ( unsigned int i = 0; i < 200; i++ ) {
		EnumChildWindows( TCmdWindow, _FindTabWindowProc, (LPARAM) &Data );

		if ( Data.FoundWindow )
			break;

		Sleep( 10 );
	}

	SetWindowText( MyConsole, OldTitle );
	return Data.FoundWindow;
}



int RefreshWindows()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"";
	HWND TCmdWindow						= NULL;
	HWND ViewWindow						= NULL;
	HWND TabWindow						= NULL;
	int k								= 0;


	if ( IsWindowVisible( ConsoleWindow ) ) {
		RedrawWindow( ConsoleWindow, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW | RDW_ALLCHILDREN );
		k++;
	}

	if ( TccVerMajor < 28 )
		return k;


	TakeCommandIPC( L"COLOR", temp );

	TakeCommandIPC( L"HWND", temp );

	if ( iswdigit( temp[0] ) )
		TCmdWindow = (HWND) wcstoul( temp, NULL, 10 );

	if ( !TCmdWindow )
		return k;
	else if ( !IsWindowVisible( TCmdWindow ) )
		return k;


	//		Try to find our tab window the fast way :

	temp[0] = '\0';
	TakeCommandIPC( L"HVIEW", temp );

	if ( iswdigit( temp[0] ) )
		ViewWindow = (HWND) wcstoul( temp, NULL, 10 );

	if ( ViewWindow ) {
		TabWindow = GetParent( ViewWindow );

		if ( ( TabWindow == ViewWindow ) || ( TabWindow == TCmdWindow ) )
			TabWindow = NULL;
	}


	//		If that didn't work, try a brute-force approach :

	if ( !TabWindow )
		TabWindow = FindTabWindow( TCmdWindow );


	if ( !TabWindow )
		return k;
	else if ( !IsWindowVisible( TabWindow ) )
		return k;

	if ( !RedrawWindow( TabWindow, NULL, NULL, RDW_INVALIDATE | RDW_UPDATENOW | RDW_ALLCHILDREN ) )
		BruteForceRefresh();

	k++;
	return k;
}



int ShowTestPattern()
{
	CONSOLE_SCREEN_BUFFER_INFO BufferInfo;
	WORD Attr		= 0;


	BufferInfo.wAttributes = 0x0007;
	GetConsoleScreenBufferInfo( ConsoleHandle, &BufferInfo );
	Attr = BufferInfo.wAttributes & 0x00ff;

	_cputws( L"\n" );

	for ( unsigned int bg = 0; bg < 16; bg++ ) {
		SetConsoleTextAttribute( ConsoleHandle, Attr );
		_cputws( L"     " );

		SetConsoleTextAttribute( ConsoleHandle, ( bg << 4 ) );
		_cputws( L"    " );

		for ( unsigned int fg = 0; fg < 16; fg++ ) {
			SetConsoleTextAttribute( ConsoleHandle, ( bg << 4 ) | fg );
			_cwprintf( L" %02X ", ( bg << 4 ) | fg );
		}

		_cputws( L"    " );
		SetConsoleTextAttribute( ConsoleHandle, Attr );
		_cputws( L"\n" );
	}

	_cputws( L"\n" );
	return 0;
}



int FormatColorString( COLORREF ColorRef, LPTSTR outString, unsigned int Format )
{
	if ( !outString )
		return -1;


	unsigned int Color		= RGB2ColorRef( ColorRef );


	switch ( Format ) {

	case 1 :

		if ( ( ( Color & 0x00f0f0f0 ) >> 4 ) == ( Color & 0x000f0f0f ) ) {
			Color = ( ( Color & 0x000f0000 ) >> 8 ) | ( ( Color & 0x00000f00 ) >> 4 ) | ( Color & 0x0000000f );
			wsprintf( outString, L"#%03x", Color );
		}
		else
			wsprintf( outString, L"#%06x", Color );

		break;

	case 2 :

		W3CColorName( Color, outString, SHORT_BUF_LEN, FALSE );
		break;

	case 3 :

		W3CColorName( Color, outString, SHORT_BUF_LEN, TRUE );
		break;

	case 4 :

		wsprintf( outString, L"%u,%u,%u", GetRValue( ColorRef ), GetGValue( ColorRef ), GetBValue( ColorRef ) );
		break;

	case 5 :

		wsprintf( outString, L"0x%06X", ColorRef );
		break;

	default :

		wsprintf( outString, L"#%06x", Color );
	}

	return 0;
}



wchar_t* FindColorEnd( LPTSTR String, BOOL WhiteSpace )
{
	if ( !String )
		return NULL;


	wchar_t *ch		= String;


	while ( *ch ) {
		if ( WhiteSpace && iswspace( *ch ) )
			return ch;

		if ( ( *ch == ';' ) || ( *ch == '/' ) )
			return ch;

		ch++;
	}

	return NULL;
}



int ReadColorsFromNativeFile( _MMFile *File, DWORD *Colors, unsigned int *Found )
{
	if ( !File || !Colors || !Found )
		return -1;


	wchar_t Line[MAX_ARG_LEN]			= L"";
	
	wchar_t *ch					= NULL;
	wchar_t *sep1				= NULL;
	unsigned int Index			= 0;
	unsigned int Value			= 0;


	while ( File->Index < File->FileSize ) {

		if ( MMGetLine( File, Line, MAX_ARG_LEN, MMGL_FLAG_SKIP_BLANKS ) )
			break;

		TrimLine( Line );
		_wcslwr_s( Line, MAX_ARG_LEN );

		if ( wcsncmp( Line, L"color", 5 ) )
			continue;

		ch = Line + 5;
		while ( iswspace( *ch ) )
			ch++;

		if ( !iswdigit( *ch ) )
			continue;

		if ( ParseInt( ch, &Index, &ch ) )
			continue;

		if ( Index > 15 )
			continue;

		while ( iswspace( *ch ) )
			ch++;

		if ( IsColonOrEquals( *ch ) ) {
			ch++;

			while ( iswspace( *ch ) )
				ch++;
		}

		sep1 = FindColorEnd( ch, ( *ch == '#' ) );
		if ( sep1 )
			*sep1 = '\0';

		Value = W3CColorValue( ch, TRUE );
		if ( Value & 0xff000000 )
			continue;

		Colors[Index] = RGB2ColorRef( Value );
		*Found |= ( 1 << Index );
	}

	return 0;
}



int CopySectionName( LPTSTR Line, LPTSTR Section )
{
	if ( !Line || !Section )
		return -1;


	wchar_t temp[SHORT_BUF_LEN]			= L"";

	unsigned int j				= 0;
	unsigned int end			= 0;


	if ( Line[0] != '[' )
		return 0;

	if ( wcschr( Line, ']' ) != ( Line + wcslen( Line ) - 1 ) )
		return 0;


	for ( wchar_t *ch = Line + 1; *ch; ch++ ) {
		if ( *ch == ']' )
			break;

		if ( iswspace( *ch ) && !j )
			continue;

		if ( j < ( SHORT_BUF_LEN - 1 ) ) {
			temp[j++] = *ch;
			temp[j] = '\0';

			if ( !iswspace( *ch ) )
				end = j;
		}
	}


	if ( end )
		temp[end] = '\0';

	if ( temp[0] ) {
		_wcslwr_s( temp, SHORT_BUF_LEN );
		wcscpy_s( Section, SHORT_BUF_LEN, temp );
	}

	return 0;
}



int ParseIniLine( LPTSTR Line, LPTSTR ItemName, LPTSTR ItemValue )
{
	if ( !Line || !ItemName || !ItemValue )
		return -1;


	wchar_t temp[SHORT_BUF_LEN]			= L"";

	wchar_t *ch					= NULL;
	unsigned int j				= 0;
	unsigned int end			= 0;


	if ( Line[0] == '[' )
		return 0;


	for ( ch = Line; *ch; ch++ ) {
		if ( *ch == '=' )
			break;

		if ( iswspace( *ch ) && !j )
			continue;

		if ( j < ( SHORT_BUF_LEN - 1 ) ) {
			temp[j++] = *ch;
			temp[j] = '\0';

			if ( !iswspace( *ch ) )
				end = j;
		}
	}


	if ( *ch != '=' )
		return 0;

	if ( end )
		temp[end] = '\0';

	if ( temp[0] == '\0' )
		return 0;

	_wcslwr_s( temp, SHORT_BUF_LEN );

	ch++;
	wcscpy_s( ItemName, SHORT_BUF_LEN, temp );
	wcscpy_s( ItemValue, SHORT_BUF_LEN, ch );
	return 1;
}



int ReadColorsFromIniFile( _MMFile *File, DWORD *Colors, unsigned int *Found )
{
	if ( !File || !Colors || !Found )
		return -1;


	wchar_t Line[MAX_ARG_LEN]			= L"";
	wchar_t Section[SHORT_BUF_LEN]		= L"";
	wchar_t ItemName[SHORT_BUF_LEN]		= L"";
	wchar_t ItemValue[SHORT_BUF_LEN]	= L"";


	size_t len						= 0;
	wchar_t *ch						= NULL;
	unsigned int Index				= 0;
	unsigned int Red				= 0;
	unsigned int Green				= 0;
	unsigned int Blue				= 0;


	while ( File->Index < File->FileSize ) {

		if ( MMGetLine( File, Line, MAX_ARG_LEN, MMGL_FLAG_SKIP_BLANKS ) )
			break;

		TrimLine( Line );
		_wcslwr_s( Line, MAX_ARG_LEN );
		len = wcslen( Line );

		if ( ( Line[0] == '[' ) && ( wcschr( Line, ']' ) == ( Line + len - 1 ) ) )
			CopySectionName( Line, Section );
		else if ( !wcscmp( Section, IniSection ) ) {
			ParseIniLine( Line, ItemName, ItemValue );

			Index = -1;
			for ( unsigned int i = 0; i < 16; i++ )
				if ( !wcscmp( ItemName, IniValueNames[i] ) )
					Index = i;

			if ( Index < 16 ) {
				Red		= -1;
				Green	= -1;
				Blue	= -1;

				ch		= ItemValue;

				while ( iswspace( *ch ) )
					ch++;

				if ( !ParseInt( ch, &Red, &ch ) ) {
					while ( iswspace( *ch ) )
						ch++;

					if ( IsListSep( *ch ) ) {
						ch++;

						while ( iswspace( *ch ) )
							ch++;

						if ( !ParseInt( ch, &Green, &ch ) ) {

							while ( iswspace( *ch ) )
								ch++;

							if ( IsListSep( *ch ) ) {
								ch++;

								while ( iswspace( *ch ) )
									ch++;

								ParseInt( ch, &Blue, &ch );
							}
						}
					}
				}

				if ( ( Red < 256 ) && ( Green < 256 ) && ( Blue < 256 ) ) {
					Colors[Index] = RGB( Red, Green, Blue );
					*Found |= ( 1 << Index );
				}
			}
		}
	}

	return 0;
}



int ReadColorsFromXmlFile( _MMFile *File, DWORD *Colors, unsigned int *Found )
{
	if ( !File || !Colors || !Found )
		return -1;


	wchar_t TagName[SHORT_BUF_LEN]			= L"";
	wchar_t TagAttribs[SHORT_BUF_LEN]		= L"";
	wchar_t TagData[MAX_ARG_LEN]			= L"";

	int Mode				= 0;

	BOOL EndOfTag			= FALSE;

	int ch					= 0;
	unsigned int j			= 0;

	unsigned int Index		= 0;
	int ColorPart			= 0;
	double Fraction			= 0.0;
	wchar_t *ch2			= NULL;
	unsigned int Red		= 0;
	unsigned int Green		= 0;
	unsigned int Blue		= 0;


	while ( File->Index < File->FileSize ) {
		ch = MMGetChar( File );
		if ( ch < 0 )
			break;

		if ( Mode == 0 ) {

			//		Parsing data between tags :

			if ( ch == '<' ) {
				j = 0;
				TagName[j] = '\0';
				EndOfTag = FALSE;
				Mode = 1;
			}
			else if ( j < ( MAX_ARG_LEN - 1 ) ) {
				if ( iswspace( ch ) )
					ch = 0x20;

				TagData[j++] = ch;
				TagData[j] = '\0';
			}

		}
		else if ( Mode == 1 ) {

			//		Parsing tag name :

			if ( ch == '>' ) {
				j = 0;
				EndOfTag = TRUE;
				Mode = 0;
			}
			else if ( iswspace( ch ) ) {
				j = 0;
				TagAttribs[j] = '\0';
				Mode = 2;
			}
			else if ( j < ( SHORT_BUF_LEN - 1 ) ) {
				TagName[j++] = ch;
				TagName[j] = '\0';
			}

		}
		else {

			//		Parsing tag attributes :

			if ( ch == '>' ) {
				j = 0;
				EndOfTag = TRUE;
				Mode = 0;
			}
			else {
				if ( iswspace( ch ) )
					ch = 0x20;

				if ( !j && ( ch == 0x20 ) )
					continue;

				if ( j < ( SHORT_BUF_LEN - 1 ) ) {
					TagAttribs[j++] = ch;
					TagAttribs[j] = '\0';
				}
			}
		}

		if ( EndOfTag ) {

			if ( TagName[0] != '/' )
				TagData[0] = '\0';


			if ( !wcscmp( TagName, L"dict" ) ) {
				ColorPart	= 0;
				Red			= -1;
				Green		= -1;
				Blue		= -1;
			}
			else if ( !wcscmp( TagName, L"/dict" ) ) {

				if ( ( Index < 16 ) && ( Red < 256 ) && ( Green < 256 ) && ( Blue < 256 ) ) {
					Colors[Index] = RGB( Red, Green, Blue );
					*Found |= ( 1 << Index );
				}

				ColorPart	= 0;
				Index		= -1;
				Red			= -1;
				Green		= -1;
				Blue		= -1;
			}
			else if ( !wcscmp( TagName, L"/key" ) ) {

				if ( !_wcsicmp( TagData, L"Red Component" ) )
					ColorPart = 1;
				else if ( !_wcsicmp( TagData, L"Green Component" ) )
					ColorPart = 2;
				else if ( !_wcsicmp( TagData, L"Blue Component" ) )
					ColorPart = 3;
				else {
					for ( unsigned int i = 0; i < 16; i++ )
						if ( !_wcsicmp( TagData, XmlColorNames[i] ) )
							Index = i;
				}
			}
			else if ( !wcscmp( TagName, L"/real" ) && TagData[0] && ( Index < 16 ) && ColorPart ) {
				ch2 = TagData;
				while ( iswspace( *ch2 ) )
					ch2++;

				_set_errno( 0 );
				Fraction = _wtof( ch2 );

				if ( !errno && ( Fraction >= 0.0 ) && ( Fraction <= 1.0 ) ) {
					switch ( ColorPart ) {

					case 1 :

						Red = (unsigned int) ( 255 * Fraction );
						break;

					case 2 :

						Green = (unsigned int) ( 255 * Fraction );
						break;

					case 3 :

						Blue = (unsigned int) ( 255 * Fraction );
						break;
					}
				}
			}


			if ( TagName[0] == '/' )
				TagData[0] = '\0';

			EndOfTag = FALSE;
		}
	}


	return 0;
}



int ColorFileType( _MMFile *File )
{
	if ( !File )
		return -1;


	wchar_t Line[MAX_ARG_LEN]			= L"";
	wchar_t Section[SHORT_BUF_LEN]		= L"";

	wchar_t *ch				= NULL;
	size_t Start			= File->Index;
	size_t len				= 0;
	BOOL FirstLine			= TRUE;
	BOOL RightSection		= FALSE;
	unsigned int test		= 0;
	int rv					= 0;


	while ( File->Index < File->FileSize ) {

		if ( MMGetLine( File, Line, MAX_ARG_LEN, MMGL_FLAG_SKIP_BLANKS ) )
			break;

		TrimLine( Line );
		_wcslwr_s( Line, MAX_ARG_LEN );
		len = wcslen( Line );


		//	check for an xml header :

		if ( FirstLine && !wcsncmp( Line, L"<?xml", 5 ) ) {
			if ( wcsstr( Line, L"?>" ) ) {
				rv = PALETTE_FILE_TYPE_XML;
				break;
			}
		}


		//	check for an ini file section header :

		if ( Line[0] == '[' ) {
			if ( wcschr( Line, ']' ) == ( Line + len - 1 ) ) {
				CopySectionName( Line, Section );
				RightSection = !wcscmp( Section, IniSection );
			}
		}
		else if ( RightSection ) {

			//	check for an ini file color value :

			test = -1;

			for ( unsigned int i = 0; i < 16; i++ ) {
				len = wcslen( IniValueNames[i] );

				if ( !wcsncmp( Line, IniValueNames[i], len ) )
					test = 1;
			}

			if ( test < 16 ) {
				ch = Line + len;

				while ( iswspace( *ch ) )
					ch++;

				if ( *ch == '=' ) {
					rv = PALETTE_FILE_TYPE_INI;
					break;
				}
			}
		}


		//	check for a native color declaration :

		if ( !wcsncmp( Line, L"color", 5 ) ) {
			ch = Line + 5;
			while ( iswspace( *ch ) )
				ch++;

			if ( iswdigit( *ch ) ) {
				if ( !ParseInt( ch, &test, &ch ) ) {
					if ( test < 16 ) {

						while ( iswspace( *ch ) )
							ch++;

						if ( IsColonOrEquals( *ch ) ) {
							ch++;

							while ( iswspace( *ch ) )
								ch++;
						}

						if ( ( *ch == '#' ) || iswalnum( *ch ) ) {
							rv = PALETTE_FILE_TYPE_NATIVE;
							break;
						}
					}
				}
			}
		}

		FirstLine = FALSE;
	}

	File->Index = Start;
	return rv;
}



int ReadColorsFromFile( LPTSTR Filename, DWORD *Colors, unsigned int Flags )
{
	if ( !Filename || !Colors )
		return -1;


	if ( !_wcsicmp( Filename, DefaultsFilename ) ) {
		for ( unsigned int i = 0; i < 16; i++ )
			Colors[i] = DefaultColors[i];

		if ( ( Flags & PALETTE_FLAG_QUIET ) == 0 )
			Printf( L"Loaded 16 colors from \"defaults\".\r\n" );

		return 0;
	}


	_MMFile File;

	BOOL Complain			= !( Flags & PALETTE_FLAG_NO_ERRORS );
	BOOL MustBeComplete		= !( Flags & PALETTE_FLAG_ALLOW_INCOMPLETE );
	BOOL ShowSuccessMsg		= !( Flags & PALETTE_FLAG_QUIET );

	int trv					= 0;
	int FileType			= PALETTE_FILE_TYPE_UNKNOWN;
	unsigned int Found		= 0;
	unsigned int Count		= 0;


	for ( unsigned int i = 0; i < 16; i++ )
		Colors[i] = -1;


	if ( trv = OpenMMFile( &File, Filename, Complain ? MM_FLAG_ERROR_MESSAGES : 0 ) ) {
		if ( Complain && ( trv == ERROR_EMPTY_FILE ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"File is empty: \"%s\"\r\n", Filename );
		}
		
		return 1;
	}


	FileType = ColorFileType( &File );

	switch ( FileType ) {

	case PALETTE_FILE_TYPE_NATIVE :

		ReadColorsFromNativeFile( &File, Colors, &Found );
		break;

	case PALETTE_FILE_TYPE_INI :

		ReadColorsFromIniFile( &File, Colors, &Found );
		break;

	case PALETTE_FILE_TYPE_XML :

		ReadColorsFromXmlFile( &File, Colors, &Found );
		break;

	default :

		CloseMMFile( &File );

		if ( Complain ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown file type: \"%s\"\r\n", Filename );
		}

		return 1;
	}

	CloseMMFile( &File );


	if ( !Found ) {
		if ( Complain ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No colors found in \"%s\"\r\n", Filename );
		}

		return 1;
	}

	if ( MustBeComplete && ( ( Found & 0xffff ) != 0xffff ) ) {
		if ( Complain ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Not all colors were found in \"%s\"\r\n", Filename );
		}

		for ( unsigned int i = 0; i < 16; i++ )
			Colors[i] = -1;

		return 1;
	}


	if ( ShowSuccessMsg ) {
		wchar_t NamePart[SHORT_BUF_LEN]		= L"";
		
		_FilenamePart( Filename, NamePart );

		for ( int i = 0; i < 16; i++ )
			if ( Found & ( 1 << i ) )
				Count++;

		Printf( L"Loaded %u color%s from \"%s\".\r\n", Count, ( Count == 1 ) ? L"" : L"s", NamePart[0] ? NamePart : Filename );
	}

	return 0;
}



int WriteXmlColor( BufOut_Buffer *OutBuffer, LPCTSTR Name, DWORD Color )
{
	wchar_t temp[SHORT_BUF_LEN]		= L"";

	int rv				= 0;


	wsprintf( temp, L"\t<key>%s</key>\n\t<dict>\n", Name );
	if ( rv = BufferedOutput( OutBuffer, temp ) )
		return rv;

	if ( rv = BufferedOutput( OutBuffer, L"\t\t<key>Color Space</key>\n\t\t<string>sRGB</string>\n" ) )
		return rv;

	if ( rv = BufferedOutput( OutBuffer, L"\t\t<key>Red Component</key>\n" ) )
		return rv;

	swprintf( temp, SHORT_BUF_LEN, L"\t\t<real>%.10f</real>\n", (double) GetRValue( Color ) / 255.0 );
	if ( rv = BufferedOutput( OutBuffer, temp ) )
		return rv;

	if ( rv = BufferedOutput( OutBuffer, L"\t\t<key>Green Component</key>\n" ) )
		return rv;

	swprintf( temp, SHORT_BUF_LEN, L"\t\t<real>%.10f</real>\n", (double) GetGValue( Color ) / 255.0 );
	if ( rv = BufferedOutput( OutBuffer, temp ) )
		return rv;

	if ( rv = BufferedOutput( OutBuffer, L"\t\t<key>Blue Component</key>\n" ) )
		return rv;

	swprintf( temp, SHORT_BUF_LEN, L"\t\t<real>%.10f</real>\n", (double) GetBValue( Color ) / 255.0 );
	if ( rv = BufferedOutput( OutBuffer, temp ) )
		return rv;

	if ( rv = BufferedOutput( OutBuffer, L"\t</dict>\n" ) )
		return rv;

	return 0;
}



int SaveColorsToXmlFile( BufOut_Buffer *OutBuffer, DWORD *Colors, _SaveInfo *SaveInfo )
{
	BOOL NeedComment		= FALSE;

	int rv				= 0;


	//	write the file header :

	if ( rv = BufferedOutput( OutBuffer, L"<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" ) )
		return rv;


	//	write additional info, if any, as a comment :

	if ( SaveInfo->Name || SaveInfo->Author || SaveInfo->Url || SaveInfo->License ) {

		if ( SaveInfo->Name )
			if ( SaveInfo->Name[0] )
				NeedComment = TRUE;

		if ( SaveInfo->Author )
			if ( SaveInfo->Author[0] )
				NeedComment = TRUE;

		if ( SaveInfo->Url )
			if ( SaveInfo->Url[0] )
				NeedComment = TRUE;

		if ( SaveInfo->License )
			if ( SaveInfo->License[0] )
				NeedComment = TRUE;


		if ( NeedComment ) {
			if ( rv = BufferedOutput( OutBuffer, L"<!--\n" ) )
				return rv;

			if ( SaveInfo->Name ) {
				if ( SaveInfo->Name[0] ) {
					if ( rv = BufferedOutput( OutBuffer, L"\tName:\t\t" ) )
						return rv;

					if ( rv = BufferedOutput( OutBuffer, SaveInfo->Name ) )
						return rv;

					if ( rv = BufferedOutput( OutBuffer, L"\n" ) )
						return rv;
				}
			}

			if ( SaveInfo->Author ) {
				if ( SaveInfo->Author[0] ) {
					if ( rv = BufferedOutput( OutBuffer, L"\tAuthor:\t\t" ) )
						return rv;

					if ( rv = BufferedOutput( OutBuffer, SaveInfo->Author ) )
						return rv;

					if ( rv = BufferedOutput( OutBuffer, L"\n" ) )
						return rv;
				}
			}

			if ( SaveInfo->Url ) {
				if ( SaveInfo->Url[0] ) {
					if ( rv = BufferedOutput( OutBuffer, L"\tURL:\t\t" ) )
						return rv;

					if ( rv = BufferedOutput( OutBuffer, SaveInfo->Url ) )
						return rv;

					if ( rv = BufferedOutput( OutBuffer, L"\n" ) )
						return rv;
				}
			}

			if ( SaveInfo->License ) {
				if ( SaveInfo->License[0] ) {
					if ( rv = BufferedOutput( OutBuffer, L"\tLicense:\t" ) )
						return rv;

					if ( rv = BufferedOutput( OutBuffer, SaveInfo->License ) )
						return rv;

					if ( rv = BufferedOutput( OutBuffer, L"\n" ) )
						return rv;
				}
			}

			if ( rv = BufferedOutput( OutBuffer, L"-->\n" ) )
				return rv;
		}
	}


	//	remainder of the file header :

	if ( rv = BufferedOutput( OutBuffer, L"<!DOCTYPE plist PUBLIC \"-//Apple//DTD PLIST 1.0//EN\" \"http://www.apple.com/DTDs/PropertyList-1.0.dtd\">\n" ) )
		return rv;

	if ( rv = BufferedOutput( OutBuffer, L"<plist version=\"1.0\">\n<dict>\n" ) )
		return rv;


	//	write the color table :

	for ( unsigned int i = 0; i < 16; i++ )
		if ( rv = WriteXmlColor( OutBuffer, XmlColorNames[i], Colors[i] ) )
			return rv;

	//	fake a background color :

	if ( rv = WriteXmlColor( OutBuffer, L"Background Color", Colors[( SaveInfo->Attribs & 0x00f0 ) >> 4] ) )
		return rv;

	//	fake a foreground color :

	if ( rv = WriteXmlColor( OutBuffer, L"Foreground Color", Colors[SaveInfo->Attribs & 0x000f] ) )
		return rv;

	//	end of file :

	if ( rv = BufferedOutput( OutBuffer, L"</dict>\n</plist>\n" ) )
		return rv;

	return 0;
}



int SaveColorsToIniFile( BufOut_Buffer *OutBuffer, DWORD *Colors, _SaveInfo *SaveInfo )
{
	wchar_t temp[MAX_ARG_LEN]		= L"";

	BOOL NeedComment	= FALSE;
	int rv				= 0;


	if ( rv = BufferedOutput( OutBuffer, L"[Table]\r\n" ) )
		return rv;

	for ( unsigned int i = 0; i < 16; i++ ) {
		wsprintf( temp, L"%-14s = %u,%u,%u\r\n", IniValueNames[i], GetRValue( Colors[i] ), GetGValue( Colors[i] ), GetBValue( Colors[i] ) );
		_wcsupr_s( temp, SHORT_BUF_LEN );

		if ( rv = BufferedOutput( OutBuffer, temp ) )
			return rv;
	}

	if ( rv = BufferedOutput( OutBuffer, L"\r\n" ) )
		return rv;


	if ( SaveInfo->Name || SaveInfo->Author || SaveInfo->Url || SaveInfo->License ) {

		if ( SaveInfo->Name )
			if ( SaveInfo->Name[0] )
				NeedComment = TRUE;

		if ( SaveInfo->Author )
			if ( SaveInfo->Author[0] )
				NeedComment = TRUE;

		if ( SaveInfo->Url )
			if ( SaveInfo->Url[0] )
				NeedComment = TRUE;

		if ( SaveInfo->License )
			if ( SaveInfo->License[0] )
				NeedComment = TRUE;


		if ( NeedComment ) {
			if ( rv = BufferedOutput( OutBuffer, L"[Info]\r\n" ) )
				return rv;

			if ( SaveInfo->Name ) {
				if ( SaveInfo->Name[0] ) {
					wsprintf( temp, L"Name    = %s\r\n", SaveInfo->Name );
					if ( rv = BufferedOutput( OutBuffer, temp ) )
						return rv;
				}
			}

			if ( SaveInfo->Author ) {
				if ( SaveInfo->Author[0] ) {
					wsprintf( temp, L"Author  = %s\r\n", SaveInfo->Author );
					if ( rv = BufferedOutput( OutBuffer, temp ) )
						return rv;
				}
			}

			if ( SaveInfo->Url ) {
				if ( SaveInfo->Url[0] ) {
					wsprintf( temp, L"URL     = %s\r\n", SaveInfo->Url );
					if ( rv = BufferedOutput( OutBuffer, temp ) )
						return rv;
				}
			}

			if ( SaveInfo->License ) {
				if ( SaveInfo->License[0] ) {
					wsprintf( temp, L"License = %s\r\n", SaveInfo->License );
					if ( rv = BufferedOutput( OutBuffer, temp ) )
						return rv;
				}
			}

			if ( rv = BufferedOutput( OutBuffer, L"\r\n" ) )
				return rv;
		}
	}

	return 0;
}



int SaveColorsToNativeFile( BufOut_Buffer *OutBuffer, DWORD *Colors, _SaveInfo *SaveInfo )
{
	wchar_t temp[MAX_ARG_LEN]		= L"";

	BOOL NeedBlank		= FALSE;
	int rv				= 0;


	if ( SaveInfo->Name ) {
		if ( SaveInfo->Name[0] ) {
			wsprintf( temp, L"; Name:     %s\r\n", SaveInfo->Name );
			if ( rv = BufferedOutput( OutBuffer, temp ) )
				return rv;

			NeedBlank = TRUE;
		}
	}

	if ( SaveInfo->Author ) {
		if ( SaveInfo->Author[0] ) {
			wsprintf( temp, L"; Author:   %s\r\n", SaveInfo->Author );
			if ( rv = BufferedOutput( OutBuffer, temp ) )
				return rv;

			NeedBlank = TRUE;
		}
	}

	if ( SaveInfo->Url ) {
		if ( SaveInfo->Url[0] ) {
			wsprintf( temp, L"; URL:      %s\r\n", SaveInfo->Url );
			if ( rv = BufferedOutput( OutBuffer, temp ) )
				return rv;

			NeedBlank = TRUE;
		}
	}

	if ( SaveInfo->License ) {
		if ( SaveInfo->License[0] ) {
			wsprintf( temp, L"; License:  %s\r\n", SaveInfo->License );
			if ( rv = BufferedOutput( OutBuffer, temp ) )
				return rv;

			NeedBlank = TRUE;
		}
	}

	if ( NeedBlank )
		if ( rv = BufferedOutput( OutBuffer, L"\r\n" ) )
			return rv;


	for ( unsigned int i = 0; i < 16; i++ ) {
		wsprintf( temp, L"Color %-2u :  #%06x\r\n", i, RGB2ColorRef( Colors[i] ) );
		if ( rv = BufferedOutput( OutBuffer, temp ) )
			return rv;
	}

	return 0;
}



int SaveColorsToFile( LPTSTR Filename, DWORD *Colors, unsigned int Mode, unsigned int Flags, WORD Attribs, _SaveInfo *SaveInfo )
{
	if ( !Filename || !SaveInfo )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	wchar_t ErrorText[SHORT_BUF_LEN]	= L"";

	BufOut_Buffer *OutBuffer			= NULL;
	BOOL ShowErrors						= ( ( Flags & PALETTE_FLAG_NO_ERRORS ) == 0 );
	int trv								= 0;
	int rv								= 0;


	if ( ( Flags & PALETTE_FLAG_ALLOW_OVERWRITE ) == 0 ) {
		if ( QueryIsFileOrDirectory( Filename ) ) {
			if ( ShowErrors ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"File exists: \"%s\"\r\n", Filename );
			}

			return ERROR_FILE_EXISTS;
		}
	}


	OutBuffer = OpenBufferedFile( Filename, 8192, BUFOUT_FILE_UTF8, 0 );

	if ( !OutBuffer ) {
		rv = GetLastError();
		if ( !rv )
			rv = 42;

		if ( ShowErrors ) {
			GetErrorText( rv, ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error opening output file: \"%s\"\r\n", Filename );
			Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
		}

		return rv;
	}

	switch ( Mode ) {

	case PALETTE_FILE_TYPE_XML :

		rv = SaveColorsToXmlFile( OutBuffer, Colors, SaveInfo );
		break;

	case PALETTE_FILE_TYPE_INI :

		rv = SaveColorsToIniFile( OutBuffer, Colors, SaveInfo );
		break;

	default :

		rv = SaveColorsToNativeFile( OutBuffer, Colors, SaveInfo );
	}


	if ( trv = CloseBufferedFile( OutBuffer ) )
		rv = trv;

	if ( rv && ShowErrors ) {
		GetErrorText( rv, ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error saving to \"%s\" :\r\n", Filename );
		Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
	}

	if ( !rv && ( ( Flags & PALETTE_FLAG_QUIET ) == 0 ) )
		Printf( L"Saved 16 colors to \"%s\" as %s.\r\n", Filename, FormatName[Mode] );

	return rv;
}



LPTSTR FindListSep( LPTSTR String )
{
	if ( !String )
		return NULL;


	wchar_t *ch = String;

	while ( *ch ) {
		if ( IsListSep( *ch ) )
			return ch;

		ch++;
	}

	return NULL;
}



int MissingColonError( LPTSTR Arg )
{
	DisplayErrorHeader();
	_wcsupr_s( Arg, MAX_ARG_LEN );
	Qprintf( ERRHANDLE, L"Expected a colon or equals sign: \"%s\"\r\n", Arg );
	return 1;
}



DLLExports int WINAPI palette( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t Filename[MAX_FILENAME_LEN]		= L"";
	wchar_t OutFilename[MAX_FILENAME_LEN]	= L"";
	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";
	wchar_t temp[SHORT_BUF_LEN]				= L"";

	CONSOLE_SCREEN_BUFFER_INFOEX BufferInfo;

	DWORD NewColor[16];
	DWORD FileColors[16];
	unsigned int Format			= 0;
	unsigned int Flags			= 0;
	unsigned int OutFileType	= PALETTE_FILE_TYPE_UNKNOWN;
	_SaveInfo SaveInfo;
	BOOL Changes				= FALSE;
	BOOL Verbose				= FALSE;
	BOOL ListColorValues		= TRUE;
	BOOL NoWarning				= FALSE;
	BOOL ShowTest				= FALSE;
	BOOL Refresh				= FALSE;
	


	memset( &BufferInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFOEX ) );
	BufferInfo.cbSize = sizeof( CONSOLE_SCREEN_BUFFER_INFOEX );

	for ( unsigned int i = 0; i < 16; i++ ) {
		NewColor[i]		= -1;
		FileColors[i]	= -1;
	}

	memset( &SaveInfo, 0x00, sizeof( _SaveInfo ) );



	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );
	

	for ( int i = 0; i < numargs; i++ ) {
		if ( !wcsncmp( arg[i], L"/?", 2 ) )
			return ShowCmdHelp( PaletteHelpText, ( arg[i][2] == '?' ) );
		else if ( !_wcsnicmp( arg[i], L"/A", 2 ) ) {
			wchar_t *ch				= arg[i] + 2;

			if ( IsColonOrEquals( *ch ) )
				ch++;
			else
				return MissingColonError( arg[i] );

			if ( *ch )
				SaveInfo.Author = ch;
			else
				SaveInfo.Author = NULL;
		}
		else if ( !_wcsnicmp( arg[i], L"/C", 2 ) || ( ( arg[i][0] == '/' ) && iswdigit( arg[i][1] ) ) ) {
			_wcsupr_s( arg[i], MAX_ARG_LEN );

			wchar_t *ch				= arg[i] + 1;
			wchar_t *sep1			= NULL;
			unsigned int Index		= 0;
			int Color				= 0;


			if ( towupper( *ch ) == 'C' )
				ch++;

			if ( ParseInt( ch, &Index, &sep1 ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad index value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( Index > 15 ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Index must be 0 - 15: %u\r\n", Index );
				return 1;
			}

			if ( !IsColonOrEquals( *sep1 ) )
				return MissingColonError( arg[i] );

			if ( ( sep1[1] == '\0' ) || IsListSep( sep1[1] ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing color name:\"%s\"\r\n", arg[i] );
				return 1;
			}

			do {
				ch = sep1 + 1;

				sep1 = FindListSep( ch );
				if ( sep1 )
					*sep1 = '\0';

				if ( *ch ) {
					Color = W3CColorValue( ch, TRUE );

					if ( Color == W3C_COLOR_UNDEFINED ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Bad color name: \"%s\"\r\n", ch );
						return 1;
					}

					NewColor[Index] = RGB2ColorRef( Color );
				}

				Index++;

				if ( sep1 )
					*sep1 = ';';
			}
			while ( sep1 && ( Index < 16 ) );

			Changes = TRUE;
			ListColorValues = FALSE;
		}
		else if ( !_wcsicmp( arg[i], L"/D" ) ) {
			if ( Filename[0] ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Can\u2019t combine /R and /D\r\n" );
				return 1;
			}

			wcscpy_s( Filename, MAX_FILENAME_LEN, DefaultsFilename );
			Changes = TRUE;
			ListColorValues = FALSE;
		}
		else if ( !_wcsicmp( arg[i], L"/E" ) )
			Flags |= PALETTE_FLAG_NO_ERRORS;
		else if ( !_wcsnicmp( arg[i], L"/F", 2 ) ) {
			wchar_t *ch = arg[i] + 2;
			if ( IsColonOrEquals( *ch ) )
				ch++;

			if ( ParseInt( ch, &Format, NULL ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad format value: \"%s\"\r\n", arg[i] );
				return 1;
			}
		}
		else if ( !_wcsicmp( arg[i], L"/I" ) )
			Flags |= PALETTE_FLAG_ALLOW_INCOMPLETE;
		else if ( !_wcsnicmp( arg[i], L"/K", 2 ) ) {
			_wcsupr_s( arg[i], MAX_ARG_LEN );

			wchar_t *ch				= arg[i] + 2;
			unsigned int Index		= 0;
			unsigned int Red		= -1;
			unsigned int Green		= -1;
			unsigned int Blue		= -1;


			if ( ParseInt( ch, &Index, &ch ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad index value: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( Index > 15 ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Index must be 0 - 15: %u\r\n", Index );
				return 1;
			}

			if ( !IsColonOrEquals( *ch ) )
				return MissingColonError( arg[i] );

			do {
				Red = Green = Blue = -1;

				//	parse red value :

				ch++;
				if ( !iswdigit( *ch ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing red value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( ParseInt( ch, &Red, &ch ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad red value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( Red > 255 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Red value must be 0 - 255: %u\r\n", Red );
					return 1;
				}

				if ( *ch != ',' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Expected a comma: \"%s\"\r\n", arg[i] );
					return 1;
				}

				//	parse green value :

				ch++;
				if ( !iswdigit( *ch ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing green value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( ParseInt( ch, &Green, &ch ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad green value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( Green > 255 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Green value must be 0 - 255: %u\r\n", Green );
					return 1;
				}

				if ( *ch != ',' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Expected a comma: \"%s\"\r\n", arg[i] );
					return 1;
				}

				//	parse blue value :

				ch++;
				if ( !iswdigit( *ch ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing blue value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( ParseInt( ch, &Blue, &ch ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad blue value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( Blue > 255 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Blue value must be 0 - 255: %u\r\n", Blue );
					return 1;
				}

				if ( *ch && ( *ch != ';' ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Expected a semicolon: \"%s\"\r\n", arg[i] );
					return 1;
				}

				NewColor[Index++] = RGB( Red, Green, Blue );
			}
			while ( *ch && ( Index < 16 ) );

			Changes = TRUE;
			ListColorValues = FALSE;
		}
		else if ( !_wcsnicmp( arg[i], L"/L", 2 ) ) {
			wchar_t *ch				= arg[i] + 2;

			if ( IsColonOrEquals( *ch ) )
				ch++;
			else
				return MissingColonError( arg[i] );

			if ( *ch )
				SaveInfo.License = ch;
			else
				SaveInfo.License = NULL;
		}
		else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
			wchar_t *ch				= arg[i] + 2;

			if ( IsColonOrEquals( *ch ) )
				ch++;
			else
				return MissingColonError( arg[i] );

			if ( *ch )
				SaveInfo.Name = ch;
			else
				SaveInfo.Name = NULL;
		}
		else if ( !_wcsicmp( arg[i], L"/Q" ) )
			Flags |= PALETTE_FLAG_QUIET;
		else if ( !_wcsnicmp( arg[i], L"/R", 2 ) ) {
			wchar_t *ch = arg[i] + 2;
			if ( IsColonOrEquals( *ch ) )
				ch++;
			else
				return MissingColonError( arg[i] );

			if ( Filename[0] ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();

				if ( !_wcsicmp( Filename, DefaultsFilename ) )
					Qprintf( ERRHANDLE, L"Can\u2019t combine /D and /R\r\n" );
				else
					Qprintf( ERRHANDLE, L"Can\u2019t read multiple /R:files: \"%s\"\r\n", arg[i] );

				return 1;
			}

			if ( *ch ) {
				/*
				QueryTrueName( ch, Filename );

				if ( Filename[0] == '\0' ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad filename: \"%s\"\r\n", arg[i] );
					return 1;
				}
				*/

				wcscpy_s( Filename, MAX_FILENAME_LEN, ch );
			}
			else {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing filename: \"%s\"\r\n", arg[i] );
				return 1;
			}

			Changes = TRUE;
			ListColorValues = FALSE;
		}
		else if ( !_wcsnicmp( arg[i], L"/S", 2 ) ) {
			wchar_t *ch = arg[i] + 2;

			if ( OutFilename[0] ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Can\u2019t save to multiple files: \"%s\"\r\n", arg[i] );
				return 1;
			}

			OutFileType = PALETTE_FILE_TYPE_UNKNOWN;

			if ( ( towupper( *ch ) == 'N' ) || ( towupper( *ch ) == 'T' ) )
				OutFileType = PALETTE_FILE_TYPE_NATIVE, ch++;
			else if ( towupper( *ch ) == 'I' )
				OutFileType = PALETTE_FILE_TYPE_INI, ch++;
			else if ( towupper( *ch ) == 'X' )
				OutFileType = PALETTE_FILE_TYPE_XML, ch++;

			if ( IsColonOrEquals( *ch ) )
				ch++;
			else
				return MissingColonError( arg[i] );

			if ( *ch ) {
				OutFilename[0] = '\0';

				if ( !wcscmp( ch, L"-" ) )
					wcscpy_s( OutFilename, MAX_FILENAME_LEN, ch );
				else
					QueryTrueName( ch, OutFilename );

				if ( OutFilename[0] == '\0' ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad filename: \"%s\"\r\n", arg[i] );
					return 1;
				}
			}
			else {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Missing filename: \"%s\"\r\n", arg[i] );
				return 1;
			}

			if ( OutFileType == PALETTE_FILE_TYPE_UNKNOWN ) {
				wchar_t Ext[SHORT_BUF_LEN]		= L"";

				OutFileType = PALETTE_FILE_TYPE_NATIVE;

				_ExtensionPart( OutFilename, Ext );

				if ( !_wcsicmp( Ext, L".ini" ) )
					OutFileType = PALETTE_FILE_TYPE_INI;
				else if ( !_wcsicmp( Ext, L".xml" ) )
					OutFileType = PALETTE_FILE_TYPE_XML;
				else if ( !_wcsicmp( Ext, L".itermcolors" ) )
					OutFileType = PALETTE_FILE_TYPE_XML;
			}

			ListColorValues = FALSE;
		}
		else if ( !_wcsicmp( arg[i], L"/T" ) ) {
			ShowTest = TRUE;
			ListColorValues = FALSE;
		}
		else if ( !_wcsnicmp( arg[i], L"/U", 2 ) ) {
			wchar_t *ch				= arg[i] + 2;

			if ( IsColonOrEquals( *ch ) )
				ch++;
			else
				return MissingColonError( arg[i] );

			if ( *ch )
				SaveInfo.Url = ch;
			else
				SaveInfo.Url = NULL;
		}
		else if ( !_wcsicmp( arg[i], L"/V" ) )
			Verbose = TRUE;
		else if ( !_wcsicmp( arg[i], L"/W" ) )
			NoWarning = TRUE;
		else if ( !_wcsicmp( arg[i], L"/X" ) )
			Flags |= PALETTE_FLAG_ALLOW_OVERWRITE;
		else if ( !_wcsicmp( arg[i], L"/Z" ) ) {
			Refresh = TRUE;
			ListColorValues = FALSE;
		}
		else if ( arg[i][0] == '/' ) {
			_wcsupr_s( arg[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
			return 1;
		}
		else {
			_wcsupr_s( arg[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
			return 1;
		}
	}


	if ( Filename[0] )
		if ( ReadColorsFromFile( Filename, FileColors, Flags ) )
			return 2;


	if ( !GetConsoleScreenBufferInfoEx( ConsoleHandle, &BufferInfo ) ) {
		GetLastErrorText( ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error getting console screen buffer info:\r\n" );
		Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
		return 2;
	}


	if ( Changes ) {

		//	copy new colors into the color table :

		for ( unsigned int i = 0; i < 16; i++ ) {
			if ( ( NewColor[i] & 0xff000000 ) == 0 )
				BufferInfo.ColorTable[i] = NewColor[i];
			else if ( ( FileColors[i] & 0xff000000 ) == 0 )
				BufferInfo.ColorTable[i] = FileColors[i];
		}

		//	work around windows api bug :

		BufferInfo.srWindow.Right += 1;
		BufferInfo.srWindow.Bottom += 1;

		//	update the console colors :

		if ( !SetConsoleScreenBufferInfoEx( ConsoleHandle, &BufferInfo ) ) {
			GetLastErrorText( ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error setting console screen buffer info:\r\n" );
			Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
			return 1;
		}

		//	refresh the console window and/or the Take Command tab window :

		RefreshWindows();

		//	if neither is visible, print a warning message :

		if ( !IsWindowVisible( ConsoleWindow ) && !NoWarning && !ModernTCmdTab() )
			QPuts( L"\r\n \u2022 The console window is not visible; color changes will not be seen.\r\n" );
	}
	else if ( Refresh )
		RefreshWindows();


	if ( ShowTest )
		ShowTestPattern();


	if ( ListColorValues || Verbose ) {
		for ( unsigned int i = 0; i < 16; i++ ) {
			FormatColorString( BufferInfo.ColorTable[i], temp, Format );
			Printf( L"Color %-2u :  %s\r\n", i, temp );
		}
	}


	if ( OutFilename[0] ) {
		SaveInfo.Attribs = BufferInfo.wAttributes & 0x00ff;

		if ( SaveColorsToFile( OutFilename, BufferInfo.ColorTable, OutFileType, Flags, BufferInfo.wAttributes & 0x00ff, &SaveInfo ) )
			return 2;
	}


	return 0;
}



int CenterWindow2( HWND hWnd )
{
	HWND ParentWindow			= GetParent( hWnd );
	RECT MyFrame				= { 0, 0, 0, 0 };
	RECT ParentFrame			= { 0, 0, 0, 0 };

	int MyWidth					= 0;
	int MyHeight				= 0;
	int ParentWidth				= 0;
	int ParentHeight			= 0;


	if ( !ParentWindow )
		return -1;

	if ( !GetWindowRect( hWnd, &MyFrame ) )
		return 0;

	if ( IsIconic( ParentWindow ) || !IsWindowVisible( ParentWindow ) ) {
		if ( !SystemParametersInfo( SPI_GETWORKAREA, 0, &ParentFrame, 0 ) )
			return 0;
	}
	else 
		if ( !GetWindowRect( ParentWindow, &ParentFrame ) )
			return 0;

	MyWidth			= MyFrame.right - MyFrame.left;
	MyHeight		= MyFrame.bottom - MyFrame.top;
	ParentWidth		= ParentFrame.right - ParentFrame.left;
	ParentHeight	= ParentFrame.bottom - ParentFrame.top;

	if ( ( MyWidth <= 0 ) || ( MyHeight <= 0 ) || ( ParentWidth <= 0 ) || ( ParentHeight <= 0 ) )
		return 0;

	SetWindowPos( hWnd, NULL, ParentFrame.left + ( ParentWidth / 2 ) - ( MyWidth / 2 ), ParentFrame.top + ( ParentHeight / 2 ) - ( MyHeight / 2 ), 0, 0, SWP_NOSIZE | SWP_SHOWWINDOW );
	SetFocus( hWnd );
	return 1;
}



UINT_PTR CALLBACK PickColorProc( HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam )
{
	if ( Msg == WM_INITDIALOG ) {
		CHOOSECOLOR *Data					= (CHOOSECOLOR*) lParam;

		if ( Data->lCustData )
			SetWindowText( hWnd, (LPTSTR) Data->lCustData );

		CenterWindow2( hWnd );

		if ( !IsWindowVisible( GetParent( hWnd ) ) || IsIconic( GetParent( hWnd ) ) )
			ForceForegroundWindow( hWnd );
	}

	return 0;
}



DLLExports int f_pickcolor( LPTSTR String )
{
	if ( !String )
		return 1;


	CONSOLE_SCREEN_BUFFER_INFOEX BufferInfo;
	CHOOSECOLOR ChooseColorData;

	COLORREF Colors[16];
	COLORREF InitColor			= -1;
	unsigned int Index			= -1;
	unsigned int Format			= 0;
	LPTSTR Title				= NULL;

	wchar_t *ch					= String;
	wchar_t *sep1				= NULL;

	unsigned int Red			= -1;
	unsigned int Green			= -1;
	unsigned int Blue			= -1;

	int rv						= 0;


	while ( iswspace( *ch ) )
		ch++;

	//	parse the color :

	if ( iswdigit( *ch ) ) {
		//	parse red,green,blue format :

		if ( ParseInt( ch, &Red, &ch ) ) {
			_wcsupr_s( ch, MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad red value: \"%s\"\r\n", ch );
			return -1;
		}

		if ( Red > 255 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Red value must be 0 - 255: %u\r\n", Red );
			return -1;
		}

		while ( iswspace( *ch ) )
			ch++;

		if ( !IsListSep( *ch ) ) {
			_wcsupr_s( ch, MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Expected a comma: \"%s\"\r\n", ch );
			return -1;
		}
		ch++;

		while ( iswspace( *ch ) )
			ch++;

		if ( ParseInt( ch, &Green, &ch ) ) {
			_wcsupr_s( ch, MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad green value: \"%s\"\r\n", ch );
			return -1;
		}

		if ( Green > 255 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Green value must be 0 - 255: %u\r\n", Green );
			return -1;
		}

		while ( iswspace( *ch ) )
			ch++;

		if ( !IsListSep( *ch ) ) {
			_wcsupr_s( ch, MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Expected a comma: \"%s\"\r\n", ch );
			return -1;
		}
		ch++;

		while ( iswspace( *ch ) )
			ch++;

		if ( ParseInt( ch, &Blue, &ch ) ) {
			_wcsupr_s( ch, MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad blue value: \"%s\"\r\n", ch );
			return -1;
		}

		if ( Blue > 255 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Blue value must be 0 - 255: %u\r\n", Blue );
			return -1;
		}

		while ( iswspace( *ch ) )
			ch++;

		InitColor = RGB( Red, Green, Blue );
	}
	else if ( *ch == '$' ) {
		//	parse console color index :

		if ( ParseInt( ch + 1, &Index, &ch ) ) {
			_wcsupr_s( ch, MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad color index: \"%s\"\r\n", ch );
			return -1;
		}

		if ( Index > 15 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Color index must be 0 - 15: %u\r\n", Index );
			return -1;
		}

		while ( iswspace( *ch ) )
			ch++;
	}
	else if ( *ch && !IsListSep( *ch ) ) {
		//	parse color name :

		sep1 = FindListSep( ch );
		if ( sep1 )
			*sep1 = '\0';

		InitColor = W3CColorValue( ch, TRUE );

		if ( InitColor == W3C_COLOR_UNDEFINED ) {
			_wcsupr_s( ch, MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad color: \"%s\"\r\n", ch );
			return -1;
		}

		InitColor = RGB2ColorRef( InitColor );

		if ( sep1 )
			*sep1 = ',', ch = sep1;
		else
			ch = String + wcslen( String );
	}
	
	if ( *ch ) {
		if ( !IsListSep( *ch ) ) {
			_wcsupr_s( ch, MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Expected a comma: \"%s\"\r\n", ch );
			return -1;
		}

		ch++;

		while ( iswspace( *ch ) )
			ch++;

		if ( iswdigit( *ch ) ) {
			//	parse the output format :

			if ( ParseInt( ch, &Format, &ch ) ) {
				_wcsupr_s( ch, MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad format value: \"%s\"\r\n", ch );
				return -1;
			}

			while ( iswspace( *ch ) )
				ch++;
		}
	}

	if ( *ch ) {
		if ( !IsListSep( *ch ) ) {
			_wcsupr_s( ch, MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Expected a comma: \"%s\"\r\n", ch );
			return -1;
		}

		ch++;

		Title = ch;
		TrimLine( Title );

		if ( Title[0] == '\0' )
			Title = NULL;
	}



	memset( &BufferInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFOEX ) );
	BufferInfo.cbSize = sizeof( CONSOLE_SCREEN_BUFFER_INFOEX );

	memset( &ChooseColorData, 0x00, sizeof( CHOOSECOLOR ) );
	ChooseColorData.lStructSize		= sizeof( CHOOSECOLOR );
	ChooseColorData.hwndOwner		= MyWindow();
	ChooseColorData.lpCustColors	= Colors;
	ChooseColorData.Flags			= CC_ANYCOLOR | CC_FULLOPEN | CC_RGBINIT | CC_ENABLEHOOK;
	ChooseColorData.lCustData		= (LPARAM) Title;
	ChooseColorData.lpfnHook		= PickColorProc;

	GetConsoleScreenBufferInfoEx( ConsoleHandle, &BufferInfo );

	memcpy( Colors, BufferInfo.ColorTable, 16 * sizeof( COLORREF ) );

	if ( ( InitColor & 0xff000000 ) == 0 )
		ChooseColorData.rgbResult = InitColor;
	else if ( Index < 16 )
		ChooseColorData.rgbResult = BufferInfo.ColorTable[Index];


	if ( ChooseColor( &ChooseColorData ) )
		FormatColorString( ChooseColorData.rgbResult, String, Format );
	else
		wcscpy_s( String, MAX_ARG_LEN, L"0" );

	return 0;
}



#include "PalEditCmd.cpp"



DLLExports int WINAPI f_palette( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";

	CONSOLE_SCREEN_BUFFER_INFOEX BufferInfo;

	unsigned int Index			= 0;
	unsigned int Format			= 0;


	memset( &BufferInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFOEX ) );
	BufferInfo.cbSize = sizeof( CONSOLE_SCREEN_BUFFER_INFOEX );


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );


	if ( arglen[0] ) {
		if ( ParseInt( arg[0], &Index, NULL ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad index value: \"%s\"\r\n", arg[0] );
			return -1;
		}

		if ( Index > 15 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Index must be 0 - 15: %u\r\n", Index );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing color index\r\n" );
		return -1;
	}

	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Format, NULL ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad format value: \"%s\"\r\n", arg[1] );
			return -1;
		}
	}


	if ( !GetConsoleScreenBufferInfoEx( ConsoleHandle, &BufferInfo ) ) {
		GetLastErrorText( ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error getting console screen buffer info:\r\n" );
		Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
		return -1;
	}

	FormatColorString( BufferInfo.ColorTable[Index], String, Format );
	return 0;
}



int TrimPath( LPTSTR Filename )
{
	if ( !Filename )
		return -1;


	wchar_t *sep1		= NULL;


	for ( wchar_t *ch = Filename; *ch; ch++ ) {
		if ( ( *ch == '/' ) || ( *ch == '\\' ) )
			sep1 = ch;
		else if ( ( *ch == ':' ) && ( ch == ( Filename + 1 ) ) && iswalnum( Filename[0] ) )
			sep1 = ch;
	}

	if ( sep1 )
		sep1[1] = '\0';
	else
		Filename[0] = '\0';

	return sep1 ? 1 : 0;
}



int AutoLoadColors()
{
	wchar_t Filename[MAX_FILENAME_LEN]		= L"%_cmdspec";
	const wchar_t AutoName[]				= L"Palette.auto";

	CONSOLE_SCREEN_BUFFER_INFOEX BufferInfo;

	DWORD FileColors[16];
	int rv				= 0;


	//		First look in the shell directory :

	ExpandVariables( Filename, 1 );

	TrimPath( Filename );
	wcscat_s( Filename, MAX_FILENAME_LEN, AutoName );

	//	wprintf( L"   * Try to autoload: '%s'\n", Filename );

	if ( ReadColorsFromFile( Filename, FileColors, PALETTE_FLAG_NO_ERRORS | PALETTE_FLAG_QUIET ) == 0 )
		rv = 1;


	//		If no luck there, try the .DLL's directory next :

	if ( !rv ) {
		Filename[0] = '\0';
		GetModuleFileName( MyDLLHandle, Filename, MAX_FILENAME_LEN );

		TrimPath( Filename );
		wcscat_s( Filename, MAX_FILENAME_LEN, AutoName );

		//	wprintf( L"   * Try to autoload: '%s'\n", Filename );

		if ( ReadColorsFromFile( Filename, FileColors, PALETTE_FLAG_NO_ERRORS | PALETTE_FLAG_QUIET ) == 0 )
			rv = 2;
	}


	//		If no valid file in either location, bug out :

	if ( !rv )
		return rv;


	//		We did get valid data, so update the console palette :

	memset( &BufferInfo, 0x00, sizeof( CONSOLE_SCREEN_BUFFER_INFOEX ) );
	BufferInfo.cbSize = sizeof( CONSOLE_SCREEN_BUFFER_INFOEX );

	if ( !GetConsoleScreenBufferInfoEx( ConsoleHandle, &BufferInfo ) )
		return 0;

	for ( unsigned int i = 0; i < 16; i++ )
		if ( ( FileColors[i] & 0xff000000 ) == 0 )
			BufferInfo.ColorTable[i] = FileColors[i];

	BufferInfo.srWindow.Right += 1;
	BufferInfo.srWindow.Bottom += 1;

	if ( !SetConsoleScreenBufferInfoEx( ConsoleHandle, &BufferInfo ) )
		return 0;

	RefreshWindows();
	return rv;
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
			MyDLLHandle = (HINSTANCE) hModule;
			break;

		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/palette.html";
	piInfo.pszDescription	= (LPTSTR) L"Display or change the console palette.";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	wchar_t shortbuf[SHORT_BUF_LEN]			= L"";


	ConsoleHandle = CreateFile( L"conout$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL );

	if ( !ConsoleHandle || ( ConsoleHandle == INVALID_HANDLE_VALUE ) ) {
		if ( !PipeShell() ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can\u2019t open the console buffer!\r\n" );
			Qprintf( ERRHANDLE, L"%s plugin not loaded.\r\n", PLUGIN_NAME );
		}

		return 1;
	}

	ConsoleWindow = GetConsoleWindow();

	RealGetWindowClass( ConsoleWindow, shortbuf, SHORT_BUF_LEN );

	if ( !_wcsnicmp( shortbuf, L"Pseudo", 6 ) ) {
		if ( !PipeShell() ) {
			shortbuf[0] = '\0';
			GetEnvironmentVariable( L"WT_SESSION", shortbuf, SHORT_BUF_LEN );

			if ( shortbuf[0] )
				Qprintf( ERRHANDLE, L"Windows Terminal is not supported!\r\n" );
			else
				Qprintf( ERRHANDLE, L"Pseudoconsoles are not supported!\r\n" );

			Qprintf( ERRHANDLE, L"%s plugin not loaded.\r\n", PLUGIN_NAME );
		}

		return 1;
	}


	GetTccVersion();
	AutoLoadColors();

	wcscpy_s( shortbuf, SHORT_BUF_LEN, L"%[_transient]%[_pipe]%[noloadmsg]" );
	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%u.%u.%u loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	CloseHandle( ConsoleHandle );
	return 0;
}



const int NumHelpTexts							= 4;
const wchar_t HelpFeature[NumHelpTexts][11]		= { L"palette", L"paledit", L"@palette", L"@pickcolor" };
const wchar_t *HelpText[NumHelpTexts]			= { PaletteHelpText, PalEditHelpText, F_PaletteHelpText, F_PickColorHelpText };



LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( ( inString[0] == '@' ) || ( inString[0] == '_' ) )
		return 0;

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, UsageText( HelpText[i] ) );
			return 1;
		}
	}

	return 0;
}



DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t WindowTitle[SHORT_BUF_LEN]		= L"";


	wsprintf( WindowTitle, L"%s Plugin Help", PLUGIN_NAME );

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			MessageBox( GetForegroundWindow(), HelpText[i], WindowTitle, MB_OK );
			return 1;
		}
	}

	return 0;
}
